package net.sourceforge.pmd.lang.java.ast;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.List;

public class DumpFacade extends JavaParserVisitorAdapter {

    private PrintWriter writer;
    private boolean recurse;

    public void initializeWith(Writer writer, String prefix, boolean recurse, JavaNode node) {
	this.writer = (writer instanceof PrintWriter) ? (PrintWriter) writer : new PrintWriter(writer);
	this.recurse = recurse;
	this.visit(node, prefix);
	try {
	    writer.flush();
	} catch (IOException e) {
	    throw new RuntimeException("Problem flushing PrintWriter.", e);
	}
    }

    @Override
    public Object visit(JavaNode node, Object data) {
	dump(node, (String) data);
	if (recurse) {
	    return super.visit(node, data + " ");
	} else {
	    return data;
	}
    }

    private void dump(JavaNode node, String prefix) {
	
	
	

	
	writer.print(prefix);

	
	writer.print(node.toString());

	
	
	
	
	
	

	
	String image = node.getImage();

	
	if (node instanceof ASTBooleanLiteral) {
	    image = String.valueOf(((ASTBooleanLiteral) node).isTrue());
	} else if (node instanceof ASTPrimaryPrefix) {
	    ASTPrimaryPrefix primaryPrefix = (ASTPrimaryPrefix) node;
	    if (primaryPrefix.usesSuperModifier()) {
		image = "super." + image;
	    } else if (primaryPrefix.usesThisModifier()) {
		if (image == null) {
		    image = "this";
		} else {
		    image = "this." + image;
		}
	    }
	} else if (node instanceof ASTPrimarySuffix) {
	    ASTPrimarySuffix primarySuffix = (ASTPrimarySuffix) node;
	    if (primarySuffix.isArrayDereference()) {
		if (image == null) {
		    image = "[";
		} else {
		    image = "[" + image;
		}
	    }
	}

	
	List<String> extras = new ArrayList<String>();

	
	if (node instanceof AccessNode) {
	    AccessNode accessNode = (AccessNode) node;
	    if (accessNode.isPackagePrivate()) {
		extras.add("package private");
	    }
	    if (accessNode.isPrivate()) {
		extras.add("private");
	    }
	    if (accessNode.isPublic()) {
		extras.add("public");
	    }
	    if (accessNode.isProtected()) {
		extras.add("protected");
	    }
	    if (accessNode.isAbstract()) {
		extras.add("abstract");
	    }
	    if (accessNode.isStatic()) {
		extras.add("static");
	    }
	    if (accessNode.isFinal()) {
		extras.add("final");
	    }
	    if (accessNode.isSynchronized()) {
		extras.add("synchronized");
	    }
	    if (accessNode.isNative()) {
		extras.add("native");
	    }
	    if (accessNode.isStrictfp()) {
		extras.add("strict");
	    }
	    if (accessNode.isTransient()) {
		extras.add("transient");
	    }
	}

	
	if (node instanceof Dimensionable) {
	    Dimensionable dimensionable = (Dimensionable) node;
	    if (dimensionable.isArray()) {
		StringBuilder extra = new StringBuilder("array");
		for (int i = 0; i < dimensionable.getArrayDepth(); i++) {
		    extra.append("[");
		}
		extras.add(extra.toString());
	    }
	}

	
	if (node instanceof ASTArguments) {
	    extras.add(String.valueOf(((ASTArguments) node).getArgumentCount()));
	} else if (node instanceof ASTAssignmentOperator) {
	    extras.add(((ASTAssignmentOperator) node).isCompound() ? "compound" : "simple");
	} else if (node instanceof ASTClassOrInterfaceBodyDeclaration) {
	    if (((ASTClassOrInterfaceBodyDeclaration) node).isAnonymousInnerClass()) {
		extras.add("anonymous inner class");
	    }
	    if (((ASTClassOrInterfaceBodyDeclaration) node).isEnumChild()) {
		extras.add("enum child");
	    }
	} else if (node instanceof ASTBlock) {
	    if (((ASTBlock) node).containsComment()) {
		extras.add("contains comment");
	    }
	} else if (node instanceof ASTClassOrInterfaceDeclaration) {
	    extras.add(((ASTClassOrInterfaceDeclaration) node).isInterface() ? "interface" : "class");
	    if (((ASTClassOrInterfaceDeclaration) node).isNested()) {
		extras.add("nested");
	    }
	} else if (node instanceof ASTConditionalExpression) {
	    if (((ASTConditionalExpression) node).isTernary()) {
		extras.add("ternary");
	    }
	} else if (node instanceof ASTConstructorDeclaration) {
	    extras.add(String.valueOf(((ASTConstructorDeclaration) node).getParameterCount()));
	    if (((ASTConstructorDeclaration) node).containsComment()) {
		extras.add("contains comment");
	    }
	} else if (node instanceof ASTExplicitConstructorInvocation) {
	    extras.add(String.valueOf(((ASTExplicitConstructorInvocation) node).getArgumentCount()));
	    if (((ASTExplicitConstructorInvocation) node).isThis()) {
		extras.add("this");
	    }
	    if (((ASTExplicitConstructorInvocation) node).isSuper()) {
		extras.add("super");
	    }
	} else if (node instanceof ASTFormalParameter) {
	    if (((ASTFormalParameter) node).isVarargs()) {
		extras.add("varargs");
	    }
	} else if (node instanceof ASTFormalParameters) {
	    extras.add(String.valueOf(((ASTFormalParameters) node).getParameterCount()));
	} else if (node instanceof ASTIfStatement) {
	    if (((ASTIfStatement) node).hasElse()) {
		extras.add("has else");
	    }
	} else if (node instanceof ASTImportDeclaration) {
	    if (((ASTImportDeclaration) node).isImportOnDemand()) {
		extras.add("on demand");
	    }
	    if (((ASTImportDeclaration) node).isStatic()) {
		extras.add("static");
	    }
	} else if (node instanceof ASTInitializer) {
	    extras.add(((ASTInitializer) node).isStatic() ? "static" : "nonstatic");
	} else if (node instanceof ASTLiteral) {
	    if (((ASTLiteral) node).isCharLiteral()) {
		extras.add("char style");
	    }
	    if (((ASTLiteral) node).isIntLiteral()) {
		extras.add("int style");
	    }
	    if (((ASTLiteral) node).isFloatLiteral()) {
		extras.add("float style");
	    }
	    if (((ASTLiteral) node).isStringLiteral()) {
		extras.add("String style");
	    }
	} else if (node instanceof ASTResultType) {
	    if (((ASTResultType) node).isVoid()) {
		extras.add("void");
	    }
	    if (((ASTResultType) node).returnsArray()) {
		extras.add("returns array");
	    }
	} else if (node instanceof ASTSwitchLabel) {
	    if (((ASTSwitchLabel) node).isDefault()) {
		extras.add("default");
	    }
	} else if (node instanceof ASTTryStatement) {
	    if (((ASTTryStatement) node).hasFinally()) {
		extras.add("has finally");
	    }
	}

	
	if (image != null || !extras.isEmpty()) {
	    writer.print(":");
	    if (image != null) {
		writer.print(image);
	    }
	    for (String extra : extras) {
		writer.print("(");
		writer.print(extra);
		writer.print(")");
	    }
	}

	writer.println();
    }
}
